/*
	Copyright (c) 2004 Jonatan Kronqvist

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
	IN THE SOFTWARE.
*/


#include "foo_bluetooth_ctrl.h"
#include "buffer.h"



bluetooth_control::bluetooth_control()
{
	console::info("Initializing");
	conn = new bluetooth_connection();
	m_activated = false;
	m_io_thread = NULL;
}


bluetooth_control::~bluetooth_control()
{
	console::info("Terminating");
	set_active(false);	// deactivate
	SAFE_DELETE(conn);
}


void bluetooth_control::set_active(bool activate)
{
	m_activated = activate;
	if(activate) {
		// Activate == run the component
		// create a window for us to send messages to
		uWNDCLASS wc;
		memset(&wc,0,sizeof(wc));
		wc.style = 0;
		wc.lpfnWndProc = g_wndproc;
		wc.hInstance = core_api::get_my_instance();
		wc.lpszClassName = "foo_bluetooth_ctrl";
		uRegisterClass(&wc);

		m_hwnd = uCreateWindowEx(0,
								"foo_bluetooth_ctrl",
								"foo_bluetooth_ctrl",
								0/*WS_OVERLAPPED|WS_CAPTION*/,
								CW_USEDEFAULT,
								CW_USEDEFAULT,
								0,0,
								core_api::get_main_window(),
								0,
								core_api::get_my_instance(),
								0);
		if(!m_hwnd) console::error("couldn't create window");

		// try to connect
		conn->set_parameter(cfg_comport.get_val());
		if(!conn->connect()) {
			console::beep();
			console::error("Could not connect!");
			console::popup();
			set_active(false);
			return;
		}
		m_io_thread = CreateThread(NULL, 0, g_read_command, reinterpret_cast<void*>(this), 0, NULL);

		console::info("Activated");
	} else {
		// Deactivate
		uDestroyWindow(m_hwnd);
		if(m_io_thread != NULL) {
			TerminateThread(m_io_thread, 0);	// FIXME: this is not very nice
			CloseHandle(m_io_thread);
			m_io_thread = NULL;
		}
		conn->disconnect();
		console::info("Deactivated");
	}
}


bool bluetooth_control::is_active()
{
	return m_activated;
}


void bluetooth_control::read_command()
{
	// all commands are 4 bytes long, which means that we can directly map
	// them to an unsigned long for instance. This very nasty hack simplifies
	// the passing of commands immensely, since we won't have to deal with (de)allocating
	// the same string in different threads
	unsigned long cmd;
	while(m_activated) {
		if(!conn->read(&cmd, 4)) {
			console::error("Read error!");
		}

		if(cmd == 0x4F464E49)	// == INFO
			Sleep(300);			// FIXME: ugly hack to allow foobar to actually start playing the song before
								// we try to read info about it and send to the phone. If we don't sleep the
								// info sent will be wrong.

		// send a message with the command, this to be sure that the command is
		// executed from the main thread to avoid deadlocks and other threading issues.
		uSendMessage(m_hwnd, WM_BT_CMD, 0, cmd);
	}
}


void bluetooth_control::send_command(unsigned long cmd)
{
	process_command(cmd);
}


/**
 * So that we don't have to have a lot of ugly #ifdefs all around
 * the code. Easier to read this way..
 */
void bluetooth_control::debug(string8 str)
{
#ifdef _DEBUG
	console::info(str);
#endif
}


string8 bluetooth_control::read_filename()
{
	string8 fname;
	unsigned short flen = 0;
	if(conn->read(&flen, 2) != 2)
		console::error("Read error!");

	flen = SWAP_SHORT(flen);
	char *filename = new char[flen];
	if(conn->read((void*)filename, flen) != flen)
		console::error("Read error!");

	fname.add_string(filename, flen);
	delete filename;
	return fname;
}


/**
 * The Bemused protocol can be found at:
 * http://www.compsoc.man.ac.uk/~ashley/bemused/BemusedProtocol.html
 */
void bluetooth_control::process_command(unsigned long _cmd)
{
	if(!core_api::assert_main_thread()) {
		console::error("process_command() called from outside of main thread.");
		return;
	}

	// FIXME: this could be done prettier by not converting back to string

	char cmd[5];
	memcpy(cmd, &_cmd, 4);

	string8 str ("Command recieved: ");
	str.add_string(cmd,4);
	debug(str);

	if(!strncmp(cmd, "CHCK", 4)) {
		conn->send("Y", 1);
	}
	else if(!strncmp(cmd, "DINF", 4)) {
		send_detailed_info();
	}
	else if(!strncmp(cmd, "DLST", 4)) {
		send_partial_list(read_filename());
	}
	else if(!strncmp(cmd, "DOWN", 4)) {
		download_file(read_filename());
	}
	else if(!strncmp(cmd, "FADE", 4)) {
		fade_stop();
	}
	else if(!strncmp(cmd, "FFWD", 4)) {
		ffwd();			// ffwd 5 seconds
	}
	else if(!strncmp(cmd, "FINF", 4)) {
		send_fileinfo(read_filename());
	}
	else if(!strncmp(cmd, "GVOL", 4)) {
		send_volume();
	}
	else if(!strncmp(cmd, "INFO", 4)) {
		send_songinfo();
	}
	else if(!strncmp(cmd, "INF2", 4)) {
		send_songinfo2();
	}
	else if(!strncmp(cmd, "LADD", 4)) {
		add_to_playlist(read_filename());
	}
	else if(!strncmp(cmd, "LIST", 4)) {
		send_entire_list();
	}
	else if(!strncmp(cmd, "NEXT", 4)) {
		next();
	}
	else if(!strncmp(cmd, "PAUS", 4)) {
		pause();
	}
	else if(!strncmp(cmd, "PLAY", 4)) {
		play(read_filename());
	}
	else if(!strncmp(cmd, "PLEN", 4)) {
		send_playlist_length();
	}
	else if(!strncmp(cmd, "PLST", 4)) {
		send_playlist();
	}
	else if(!strncmp(cmd, "PREV", 4)) {
		prev();
	}
	else if(!strncmp(cmd, "REPT", 4)) {
		char repeat=0;
		conn->read(&repeat, 1);
		set_repeat( repeat == 1 );
	}
	else if(!strncmp(cmd, "RMAL", 4)) {
		clear_playlist();
	}
	else if(!strncmp(cmd, "RWND", 4)) {
		rwnd();
	}
	else if(!strncmp(cmd, "SHFL", 4)) {
		char shuffle=0;
		conn->read(&shuffle, 1);
		set_shuffle( shuffle == 1 );
	}
	else if(!strncmp(cmd, "SEEK", 4)) {
		long pos;
		conn->read(&pos, 4);
		seek(pos);
	}
	else if(!strncmp(cmd, "SHUT", 4)) {
		// shut down PC.. not implemented
	}
	else if(!strncmp(cmd, "STEN", 4)) {
		stop_at_eof_track();
	}
	else if(!strncmp(cmd, "SLCT", 4)) {
		short idx;
		conn->read(&idx, 2);
		select_at(idx);
	}
	else if(!strncmp(cmd, "STOP", 4)) {
		stop();
	}
	else if(!strncmp(cmd, "STRT", 4)) {
		play_current();
	}
	else if(!strncmp(cmd, "VOLM", 4)) {
		char vol;
		conn->read(&vol, 1);
		set_volume(vol);
	}
	else if(!strncmp(cmd, "VERS", 4)) {
		send_version();
	}
}


void bluetooth_control::send_detailed_info()
{
	debug("Sending detailed info");

	/* format of response:
		length	field name
		7		"DINFACK"
		4		bitrate
		4		samplerate
		4		number of channels
	*/
	file_info_i_full finf;
	metadb_handle *h = play_control::get()->get_now_playing();
	h->handle_lock();
	h->handle_query(&finf);
	long bitrate = SWAP_LONG((long)finf.info_get_bitrate());

	audio_chunk_i chunk;
	input_helper in;
	in.open(h);
	in.run(&chunk);
	long samplerate = SWAP_LONG(chunk.get_srate());
	long channels = SWAP_LONG(chunk.get_channels());
	in.close();

	h->handle_unlock();
	h->handle_release();

	buffer buf(19);
	buf.append("DINFACK");
	buf.append(bitrate);
	buf.append(samplerate);
	buf.append(channels);
	conn->send(buf);
}


void bluetooth_control::send_dirlist(DirTreeNode *root)
{
	if(root == NULL) return;

	buffer buf(root->name.length()+2);
	buf.append(root->type);
	buf.append(root->name);
	buf.append('\0');
	conn->send(buf);

	string8 tmpstr;
	tmpstr.add_int(root->type >> 4);
	tmpstr.add_char(' ');
	tmpstr.add_int(root->type & 0xF);
	tmpstr.add_char(' ');
	tmpstr.add_string(root->name);
	debug(tmpstr);

	send_dirlist(root->first_child);
	for(unsigned int i=0; i<root->siblings.size(); i++)
		send_dirlist(root->siblings[i]);
}


DirTreeNode *bluetooth_control::build_dirtree(string8 root, string8 prefix, unsigned char type, bool recursive)
{
	uFindFile	*find;
	string8		filename;
	// cur is pointing to a dir with name in root
	DirTreeNode	*cur = new DirTreeNode(root, type);

	string8 path, glob;
	path.add_string(prefix);
	path.add_string(root);
	path.add_string("\\");

	glob.set_string(path);
	glob.add_string("*.*");
	find = uFindFirstFile(glob);

	// TODO: reorganise this loop..
	do {
		filename.set_string(find->GetFileName());
		if(find->IsDirectory()) {
			if(strncmp(find->GetFileName(), ".", 1)) {	// don't do anything for directories starting with "." e.g. "." and ".."
				if(recursive) {
					if(cur->first_child == NULL)
						cur->first_child = build_dirtree(filename, path, DIRLST_CHILD+DIRLST_DIR, true);
					else
						cur->siblings.push_back(build_dirtree(filename, path, DIRLST_SIBLING+DIRLST_DIR, true));
				} else {
					if(cur->first_child == NULL)
						cur->first_child = new DirTreeNode(filename, DIRLST_CHILD + DIRLST_UXDIR);
					else
						cur->siblings.push_back(new DirTreeNode(filename, DIRLST_SIBLING + DIRLST_UXDIR));
				}
			}
		} else {
			// normal file
			if(cur->first_child == NULL)
				cur->first_child = new DirTreeNode(filename, DIRLST_CHILD + DIRLST_FILE);
			else
				cur->siblings.push_back(new DirTreeNode(filename, DIRLST_SIBLING + DIRLST_FILE));
		}
	} while(find->FindNext());

	if(cur->siblings.empty()) {		// last is a child, mark as only child
		if(cur->first_child != NULL)
			cur->first_child->type = DIRLST_OCHILD + (cur->first_child->type & 0x0F);
	} else {	// last is a sibling, mark as last sibling
		cur->siblings.back()->type = DIRLST_LSIBLING + (cur->siblings.back()->type & 0x0F);
	}

	delete find;

	return cur;
}


void bluetooth_control::delete_dirtree(DirTreeNode *root)
{
	if(root->first_child != NULL)
		delete_dirtree(root->first_child);

	for(unsigned int i=0; i<root->siblings.size(); i++) {
		delete_dirtree(root->siblings[i]);
	}
	root->siblings.clear();

	SAFE_DELETE(root);
}


void bluetooth_control::download_file(string8 fname) {}	// NOT IMPLEMENTED


void bluetooth_control::fade_stop()
{
	// TODO: fix the actual fading :)
	stop();
}


void bluetooth_control::ffwd()		// 5 secs
{
	// TODO: figure out if the param is in seconds
	// also, where the hell in the client can we seek? :)
	play_control::get()->playback_seek_delta(5);
}


void bluetooth_control::rwnd()		// 5 secs
{
	// TODO: figure out if the param is in seconds
	// also, where the hell in the client can we seek? :)
	play_control::get()->playback_seek_delta(-5);
}


void bluetooth_control::send_fileinfo(string8 fname)
{
	long size = SWAP_LONG(4096);
	
	file_info_i_full finf;
	metadb_handle *h = play_control::get()->get_now_playing();
	h->handle_lock();
	h->handle_query(&finf);

	finf.get_file_path();
	uFindFile *find = uFindFirstFile(finf.get_file_path());
	size = SWAP_LONG((long)find->GetFileSize());
	delete find;

	h->handle_unlock();
	h->handle_release();

	buffer buf(11);
	buf.append("FINFACK");
	buf.append(size);
	conn->send(buf);
}


void bluetooth_control::send_volume()
{
	/* format of response:
		length	field name
		7		"GVOLACK"
		1		volume
	*/

	int dBvol = play_control::get()->get_volume();	// in 0.01 dB increments
	unsigned char vol = (unsigned char)(256.0f / (dBvol/100.0f+100.0f));	// loose some precision -> -100 .. 0 dB

	string8 tmpstr ("Sending vol: ");
	tmpstr.add_int(vol);
	tmpstr.add_char(',');
	tmpstr.add_int(dBvol);
	console::info(tmpstr);

	buffer buf(8);
	buf.append("VOLACK");
	buf.append(vol);
	conn->send(buf);
}


void bluetooth_control::set_volume(unsigned char vol)
{
	// NOTE: this is closely knit to send_volume!!
	int dBvol = int((vol/256.0f * 10000.0f) - 10000.0f);	// 0.01 dB increments
	// because of loss of precision:
	if(vol == 255) dBvol = 0;
	play_control::get()->set_volume(dBvol);
}


void bluetooth_control::send_songinfo()
{
	debug("Sending songinfo");

	/* format of the response:
		length	field name
		7 		"INFOACK"
		1		play state
		4		song length in seconds
		4		current time played in seconds
		1		shuffle state
		1		repeat state
	*/
	char playstate = 0;
	if(play_control::get()->is_playing()) playstate = playstate | 0x1;
	if(play_control::get()->is_paused()) playstate = playstate | 0x2;
	long playlen = SWAP_LONG((long)play_control::get()->playback_get_length());
	long playtime = SWAP_LONG((long)play_control::get()->get_playback_time());
	char shuffle = 0, repeat = 0;
	string8 tmpstr;
	playback_order_helper::get_config_by_name(tmpstr);
	if(tmpstr == "Random") shuffle = 1;
	if(tmpstr == "Repeat") repeat = 1;

	string8 name;
	play_control::get()->playback_format_title(name, cfg_playlist_fmt);
	name.add_char('\0');

	buffer buf(18+name.length());
	buf.append("INFOACK");
	buf.append(playstate);
	buf.append(playlen);
	buf.append(playtime);
	buf.append(shuffle);
	buf.append(repeat);
	buf.append(name);
	conn->send(buf);
}


void bluetooth_control::send_songinfo2()
{
	debug("Sending songinfo2");
	send_songinfo();
	conn->send("\0", 1);
}


void bluetooth_control::add_to_playlist(string8 fname)
{
	string8 url ("file://");
	url.add_string(fname);
	debug(url);
	playlist_oper::get()->add_location(url);
}


void bluetooth_control::send_entire_list()
{
	string8 rootdir (cfg_musicdir);
	DirTreeNode *root = build_dirtree(rootdir, "", DIRLST_ROOT+DIRLST_DIR, true);

	conn->send("LISTACK", 7);
	send_dirlist(root);

	unsigned char terminator[2];
	terminator[0] = 0xFF;
	terminator[1] = 0x00;
	conn->send(terminator, 2);		// send terminating node

	delete_dirtree(root);
}


void bluetooth_control::send_partial_list(string8 dir)
{
	DirTreeNode *root = build_dirtree(dir, "", DIRLST_ROOT+DIRLST_DIR, false);

	conn->send("LISTACK", 7);
	send_dirlist(root);

	unsigned char terminator[2];
	terminator[0] = 0xFF;
	terminator[1] = 0x00;
	conn->send(terminator, 2);		// send terminating node

	delete_dirtree(root);
}


void bluetooth_control::next()
{
	play_control::get()->play_start(play_control::TRACK_COMMAND_NEXT);
}


void bluetooth_control::prev()
{
	play_control::get()->play_start(play_control::TRACK_COMMAND_PREV);
}


void bluetooth_control::pause()
{
	play_control::get()->pause();
}


void bluetooth_control::stop()
{
	play_control::get()->play_stop();
}


void bluetooth_control::play(string8 fname)
{
	string8 url ("file://");
	url.add_string(fname);
	debug(url);
	playlist_oper::get()->add_location(url, true);							// add to the end of playlist
	playlist_oper::get()->play_item(playlist_oper::get()->get_count()-1);	// play the last entry
}


void bluetooth_control::play_current()
{
	play_control::get()->play_start();
}


void bluetooth_control::send_playlist_length()
{
	// FIXME: this may overflow, should we add a check
	// and report the error to user or "fix" the protocol and client?
	short playlist_len = (short)playlist_oper::get()->get_count();
	playlist_len = SWAP_SHORT(playlist_len);
	conn->send(&playlist_len, 2);
}


void bluetooth_control::send_playlist()
{
	debug("Sending playlist");
	conn->send("PLSTACK", 7);
	// FIXME: possible overflow (if > 65536 songs in playlist)
	unsigned short cur_idx = SWAP_SHORT((unsigned short)playlist_oper::get()->get_now_playing());
	conn->send(&cur_idx, 2);

	metadb_handle *handle;
	unsigned short playlist_len = (unsigned short)playlist_oper::get()->get_count();
	for(int i=0; i<playlist_len; i++) {
		handle = playlist_oper::get()->get_item(i);
		string8 title;
		handle->handle_format_title(title, cfg_playlist_fmt.get_val(), NULL);
		title.add_char('\n');
		conn->send((void*)title.get_ptr(), title.length());
		handle->handle_release();
	}
	conn->send("\0", 1);
	debug("Playlist sent");
}


void bluetooth_control::select_at(short idx)
{
	// big endian -->
	idx = SWAP_SHORT(idx);

	string8 str ("Selected track for play at: ");
	str.add_int(idx);
	debug(str);
	
	playlist_oper::get()->set_sel_one(idx);
	playlist_oper::get()->play_item(idx);
}


void bluetooth_control::clear_playlist()
{
	playlist_oper::get()->remove_all();
}


void bluetooth_control::set_repeat(bool repeat)
{
	if(repeat)
		playback_order_helper::set_config_by_name("Repeat");
	else
		playback_order_helper::set_config_by_name("Default");
}


void bluetooth_control::set_shuffle(bool shuffle)
{
	if(shuffle)
		playback_order_helper::set_config_by_name("Random");
	else
		playback_order_helper::set_config_by_name("Default");
}


void bluetooth_control::seek(long pos)
{
	pos = SWAP_LONG(pos);
	if(play_control::get()->playback_can_seek())
		play_control::get()->playback_seek((double)pos);
}


void bluetooth_control::stop_at_eof_track()
{
	play_control::get()->set_stop_after_current(true);
}


void bluetooth_control::send_version()
{
	debug("Sending version info to phone.");

	unsigned char ver[2];
	ver[0] = 1;
	ver[1] = 73;
	conn->send(ver, 2);
}






/*******************
 * Global functions
 *******************/

static DWORD WINAPI g_read_command(LPVOID param)
{
	bluetooth_control *ctrl = reinterpret_cast<bluetooth_control*>(param);
	ctrl->read_command();
	return 0;
}

LRESULT CALLBACK g_wndproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_BT_CMD:
		g_bt_ctrl->send_command(lParam);
		break;
	}
	return uDefWindowProc(hwnd,msg,wParam,lParam);
}





/****************
 * CONFIG dialog
 ****************/
BOOL CALLBACK config_bluetooth_ctrl::ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	string8 tmpstr,tmp2str;
	switch(msg)
	{
	case WM_INITDIALOG:
		uSetDlgItemText(wnd, IDC_MUSICDIR, cfg_musicdir);
		uSetDlgItemText(wnd, IDC_COMPORT, cfg_comport);
		uSetDlgItemText(wnd, IDC_PLAYLIST_FMT, cfg_playlist_fmt);
		uSendDlgItemMessage(wnd,IDC_AUTOCONNECT,BM_SETCHECK,cfg_autoconnect.get_value(),0);
		EnableWindow(wnd, TRUE);
		break;

	case WM_DESTROY:
		break;

	case WM_COMMAND:
		switch(wp) {
		case IDC_COMPORT | (EN_CHANGE<<16):
			uGetDlgItemText(wnd, IDC_COMPORT, tmpstr);
			uStringUpper(tmp2str, tmpstr.get_ptr(), tmpstr.length());
			cfg_comport = tmp2str;
			break;

		case IDC_CONNECT_BUTTON:
			uGetDlgItemText(wnd, IDC_COMPORT, tmpstr);
			uStringUpper(tmp2str, tmpstr.get_ptr(), tmpstr.length());
			cfg_comport = tmp2str;
			// reconnect
			g_bt_ctrl->set_active(false);
			g_bt_ctrl->set_active(true);
			break;

		case IDC_MUSICDIR_BUTTON:
			uGetDlgItemText(wnd,IDC_MUSICDIR,tmpstr);
			if (uBrowseForFolder(wnd,"Select music directory",tmpstr))
			{
				uSetDlgItemText(wnd,IDC_MUSICDIR,tmpstr);
				cfg_musicdir = tmpstr;
			}			
			break;

		case IDC_MUSICDIR | (EN_CHANGE<<16):
			uGetDlgItemText(wnd, IDC_MUSICDIR, tmpstr);
			cfg_musicdir = tmpstr;
			break;

		case IDC_AUTOCONNECT:
			cfg_autoconnect.set_value(uSendMessage((HWND)lp, BM_GETCHECK, 0, 0));
			break;

		case IDC_PLAYLIST_FMT | (EN_CHANGE<<16):
            uGetDlgItemText(wnd, IDC_PLAYLIST_FMT, tmpstr);
			cfg_playlist_fmt = tmpstr;
			break;

		case IDC_TAGHELP:
			menu_manager::run_command("Foobar2000/Titleformatting help");
			break;
		}
		break;
	}
	return 0;
}
