/*
	Copyright (c) 2004 Jonatan Kronqvist

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
	IN THE SOFTWARE.
*/

/*
	Based on the Bemused protocol
	http://www.compsoc.man.ac.uk/~ashley/bemused/BemusedProtocol.html
	Client available from the same location.
*/


#ifndef __FOO_BLUETOOTH_CTRL_H__
#define __FOO_BLUETOOTH_CTRL_H__

#include "../SDK/foobar2000.h"
#include "../helpers/helpers.h"

#include <vector>

#include "resource.h"
#include "connection.h"
#include "macros.h"



#define WM_BT_CMD	0x8000


static DWORD WINAPI g_read_command(LPVOID param);
LRESULT CALLBACK g_wndproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);




/**
 * config variables
 */
static cfg_string cfg_comport("bluetooth_ctrl_comport", "COM4");
static cfg_string cfg_musicdir("bluetooth_ctrl_musicdir", "H:\\mp3");
static cfg_string cfg_playlist_fmt("bluetooth_ctrl_playlist_fmt", "$if2(%artist% - %title%, %_filename_ext%)");
static config_var_int_i cfg_autoconnect("bluetooth_ctrl_autoconnect", 0);



/***************************
 * Directory Tree structure
 ***************************/
// upper nibble
#define DIRLST_ROOT		(0x0<<4)
#define DIRLST_CHILD	(0x1<<4)
#define DIRLST_OCHILD	(0x2<<4)
#define DIRLST_SIBLING	(0x3<<4)
#define DIRLST_LSIBLING	(0x4<<4)
#define DIRLST_TERM		(0xF<<4)
// lower nibble
#define DIRLST_DRIVE	0x0
#define DIRLST_DIR		0x1
#define DIRLST_FILE		0x2
#define DIRLST_UXDIR	0x3

class DirTreeNode
{
public:
	unsigned char				type;	// upper and lower nibbles
	DirTreeNode*				first_child;
	std::vector<DirTreeNode*>	siblings;
	string8						name;
	DirTreeNode(string8 n, unsigned char t) {
		name = n;
		type = t;
		first_child = NULL;
	}
};



/*************
 * Main class
 *************/
class bluetooth_control
{
public:
	bluetooth_control();
	~bluetooth_control();

	void set_active(bool activate);
	bool is_active();
	void read_command();
	void send_command(unsigned long cmd);

private:
	connection *conn;
	bool	m_activated;
	metadb	*m_mdb;
	HANDLE	m_io_thread;
	HWND	m_hwnd;


	void debug(string8 str);

	/**
	 * The Bemused protocol can be found at:
	 * http://www.compsoc.man.ac.uk/~ashley/bemused/BemusedProtocol.html
	 */
	void process_command(unsigned long _cmd);
	string8 read_filename();

	DirTreeNode* build_dirtree(string8 root, string8 prefix, unsigned char type, bool recursive);
	void delete_dirtree(DirTreeNode *root);
	void send_detailed_info();
	void send_dirlist(DirTreeNode *root);
	void download_file(string8 fname);
	void fade_stop();
	void ffwd();		// 5 secs
	void rwnd();		// 5 secs
	void send_fileinfo(string8 fname);
	void send_volume();
	void set_volume(unsigned char vol);
	void send_songinfo();
	void send_songinfo2();
	void add_to_playlist(string8 fname);
	void send_entire_list();
	void send_partial_list(string8 dir);
	void next();
	void prev();
	void pause();
	void stop();
	void play(string8 fname);
	void play_current();
	void send_playlist_length();
	void send_playlist();
	void select_at(short idx);
	void clear_playlist();
	void set_repeat(bool repeat);
	void set_shuffle(bool shuffle);
	void seek(long pos);
	void stop_at_eof_track();
	void send_version();
};

bluetooth_control *g_bt_ctrl = NULL;




/**
 * Called on foobar init/quit
 */
class initquit_bluetooth : public initquit
{
public:
	virtual void on_init()
	{ 
		g_bt_ctrl = new bluetooth_control();	// create new instance
		if(cfg_autoconnect.get_value() != 0)
			g_bt_ctrl->set_active(true);				// activate if autoconnect is set
	}
	virtual void on_quit()
	{
		g_bt_ctrl->set_active(false);
		SAFE_DELETE(g_bt_ctrl);
	}
};




/**
 * Components-Toggle Bluetooth Connection menu item
 */
class components_menu_item_bluetooth : public menu_item_main_single
{
public:
	void get_name(string_base &out)
	{
		out.set_string("Components/Bluetooth Connection");
	}
	void run()
	{
		g_bt_ctrl->set_active(!g_bt_ctrl->is_active());		// toggle active
	}
	bool is_checked()
	{
		return g_bt_ctrl->is_active();
	}
	virtual bool get_description(string_base & out) {return false;}
};




/**
 * Config menu (preferences)
 */
class config_bluetooth_ctrl : public config
{
	static BOOL CALLBACK ConfigProc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);
	HWND create(HWND parent)
	{
		return uCreateDialog(IDD_CONFIG_BLUETOOTH_CTRL,parent,ConfigProc);
	}
	const char * get_name() {return "Bluetooth Remote Control";}
	const char * get_parent_name() {return "Components";}
};




static service_factory_single_t<initquit, initquit_bluetooth> iqbt;
static service_factory_single_t<menu_item, components_menu_item_bluetooth> cmibt;
static service_factory_single_t<config, config_bluetooth_ctrl> cbtc;




DECLARE_COMPONENT_VERSION(
	"Bluetooth Remote Control",
	"0.3",
	"Copyright (c) 2004 Jonatan Kronqvist <jonatan@oxygenite.org>"
);


#endif /*__FOO_BLUETOOTH_CTRL_H__*/